/* Text Printer class implementation *\
\* By Justin Pease                   */

#include "gamex-text.h"
#include "gamex-camera.h"
#include "gamex-image.h"
#include "gamex-win-dx.h"

// TextPrinter - Initialize
// Load text font image
void TextPrinter::Initialize(ImageX * alphabetImage, int charsAcrossImage, const char * charListInImageOrder)
{
	assert(alphabetImage != NULL);
	assert(charsAcrossImage >= 1);

	z_charPixels = alphabetImage->GetWidth() / charsAcrossImage;
	z_fCharPixels = (float)alphabetImage->GetWidth() / (float)charsAcrossImage;

	assert(z_charPixels > 3); // make sure the letters aren't ridiculously tiny

	text_img = alphabetImage;

	int numChars = (int)strlen(charListInImageOrder);

	int totalTop = 0;
	int totalBot = 0;
	int total = 0;

	GameX.AccessPixels(text_img);

	// construct lookup info for each character
	for(int i = 0 ; i < 256 ; i++)
	{
		char c = (char)i;
		int xs = -1, ys = -1;

		z_chars[i].visible = false;
		for(int j = 0 ; j < numChars ; j++)
		{
			if(charListInImageOrder[j] == c) // if it's the current character
			{
				xs = j % charsAcrossImage;
				ys = j / charsAcrossImage;
				z_chars[i].visible = true;
				break;
			}
		}

		z_chars[i].xs = xs*z_charPixels;
		z_chars[i].ys = ys*z_charPixels;

		// initialize min and max ranges to max and min possible, respectively
		for(int interval = 0 ; interval < z_HEIGHT_LEVELS ; interval++)
		{
			z_chars[i].xl[interval] = 255;
			z_chars[i].xr[interval] = 0;
		}

		if(z_chars[i].visible)
		{
			// now for the hard part: sample the image for the character's horizontal extents at each height range

			int r,g,b;
			int minX = z_chars[i].xs, maxX = z_chars[i].xs + z_charPixels;
			int minY = z_chars[i].ys, maxY = z_chars[i].ys + z_charPixels;

			if(minX < 0 || maxX < 0 || minY < 0 || maxY < 0 || minX > text_img->GetWidth() || maxX > text_img->GetWidth() || minY > text_img->GetHeight() || maxY > text_img->GetHeight())
			{
				z_chars[i].visible = false;
				continue;
			}

			assert(minX >= 0 && minY >= 0 && maxX <= text_img->GetWidth() && maxY <= text_img->GetHeight());
			bool findHeights = (c >= 'A' && c <= 'Z');
			bool foundTop = false;
			int botSoFar = 0;


			for(int y = minY ; y < maxY ; y++)
			{
				int interval = z_HEIGHT_LEVELS * (y-minY) / (maxY-minY);
				assert(interval >= 0 && interval < z_HEIGHT_LEVELS);
				bool foundSomething = false;

				// go from left to right, set min when we find something
				for(int x = minX ; x < maxX ; x++)
				{
					GameX.ReadPixel(x,y, r,g,b);
					b = max(r,max(g,b));

					if(b > GameX.z_win_maskming)
					{
						if(!foundTop && findHeights)
						{
							totalTop += y-minY;
							total++;
							foundTop = true;
						}
						botSoFar = max(botSoFar, y);

						int pos = (x - minX) * 256 / z_charPixels;
						assert(pos >= 0 && pos < 256);

						if(z_chars[i].xl[interval] > pos)
							z_chars[i].xl[interval] = pos;
						foundSomething = true;
						break;
					}
				}

				// go from right to left, set max when we find something
				if(foundSomething)
				for(int x = maxX-1 ; x >= minX ; x--)
				{
					GameX.ReadPixel(x,y, r,g,b);
					b = max(r,max(g,b));

					if(b > GameX.z_win_maskming)
					{
						int pos = (x - minX) * 256 / z_charPixels;
						assert(pos >= 0 && pos < 256);

						if(z_chars[i].xr[interval] < pos)
							z_chars[i].xr[interval] = pos;
						break;
					}
				}

///				if(z_chars[i].xr[interval] < 255)
///					z_chars[i].xr[interval]++; // add tiny bit of spacing
			}

			// prevent total interlocking for things like '_' and '...'
			if(z_chars[i].visible)
			{
				int xMin = 255, xMax = 0;
				for(int interval = 0 ; interval < z_HEIGHT_LEVELS ; interval++)
				{
					if(xMin > z_chars[i].xl[interval])
						xMin = z_chars[i].xl[interval];
					if(xMax < z_chars[i].xr[interval])
						xMax = z_chars[i].xr[interval];
				}
				int xAvg = (xMin + xMax) / 2;

				for(int interval = 0 ; interval < z_HEIGHT_LEVELS ; interval++)
					if(z_chars[i].xl[interval] == 255 && z_chars[i].xr[interval] == 0)
						if(c != '_')
							z_chars[i].xl[interval] = z_chars[i].xr[interval] = xAvg;
						else // special case to make _ fill whole character:
							z_chars[i].xl[interval] = xMin, z_chars[i].xr[interval] = xMax;
			}

			if(foundTop)
				totalBot += botSoFar-minY;
		}
		else
		{
			for(int interval = 0 ; interval < z_HEIGHT_LEVELS ; interval++)
			{
				z_chars[i].xl[interval] = 0;
				z_chars[i].xr[interval] = 0;
			}
		}
	}

	GameX.EndPixelAccess();

	// special case substitute for ' for  or vice versa, and same for " and 
	// (if either of those are missing and their nearly-equivalent partner isn't)
#define apo1 z_chars[((unsigned char)'\'')] // NOTE: the cast to (unsigned char) is absolutely necessary
#define apo2 z_chars[((unsigned char)'' )] //       because we're relying on the negative wrap-around to 255
#define quo1 z_chars[((unsigned char)'\"')]
#define quo2 z_chars[((unsigned char)'' )]
	if(!apo1.visible && apo2.visible)      apo1 = apo2;
	else if(!apo2.visible && apo1.visible) apo2 = apo1;
	if(!quo1.visible && quo2.visible)      quo1 = quo2;
	else if(!quo2.visible && quo1.visible) quo2 = quo1;

	// special case the white space since there's no image data to go by for it
	for(int interval = 0 ; interval < z_HEIGHT_LEVELS ; interval++)
	{
		z_chars[((unsigned char)' ')].xl[interval] = 0;
		z_chars[((unsigned char)' ')].xr[interval] = 64;

		z_chars[((unsigned char)'\t')].xl[interval] = 0;
		z_chars[((unsigned char)'\t')].xr[interval] = 255;

		z_chars[0].xl[interval] = 0;
		z_chars[0].xr[interval] = 255;
	}

	// store info to kern from any character to any other character (into a big table)
	for(int c = 0 ; c < 256 ; c++)
	for(int nc = 0 ; nc < 256 ; nc++)
	{
		unsigned char minPos = 0;
		for(int i = 0 ; i < z_HEIGHT_LEVELS ; i++)
		{
			if(z_chars[nc].xl[i]+minPos < z_chars[c].xr[i])
				minPos = z_chars[c].xr[i];
		}
		z_charKern[c][nc] = minPos;
	}

	z__avgTop = ((total) ? (float)totalTop / (float)total - 2 : 0) / z_fCharPixels;
	z__avgBot = ((total) ? (float)totalBot / (float)total + 2 : z_fCharPixels) / z_fCharPixels;
}

void TextPrinter::Initialize(ImageX * alphabetImage, int charsAcrossImage)
{
	Initialize(alphabetImage,charsAcrossImage,"~!@#$%^&*()_+,./`0123456789-=;:?ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz<>\\|\"'[]{}");
}

void TextPrinter::PrintTextScreenProportional(const char * str, float x, float y, float heightRatio, float xStretch, bool centered, int draw_mode, ColorX & color, float angle, bool rotate_baseline, bool fixed)
{
	if(xStretch <= 0.0f || heightRatio <= 0.0f) return;
	float font_size = (heightRatio * GameX.GetHeight()) / (z__avgBot - z__avgTop);
	float drawYTop = y - (font_size * z__avgTop);
	float scaley = 1.0f / xStretch;
	PrintText(str, x, drawYTop, font_size * xStretch, centered, draw_mode, color, angle, scaley, rotate_baseline, fixed);
}

float TextPrinter::GetVisibleStringWidthProportional(const char * str, float heightRatio, float xStretch)
{
	if(xStretch <= 0.0f || heightRatio <= 0.0f) return 0;
	float font_size = (heightRatio * GameX.GetHeight()) / (z__avgBot - z__avgTop);
	return GetVisibleStringWidth(str,0,32767,font_size * xStretch);
}

float TextPrinter::GetVisibleStringWidthProportional(const char * str, int first_index, int last_index, float heightRatio, float xStretch)
{
	if(xStretch <= 0.0f || heightRatio <= 0.0f) return 0;
	float font_size = (heightRatio * GameX.GetHeight()) / (z__avgBot - z__avgTop);
	return GetVisibleStringWidth(str,first_index,last_index,font_size * xStretch);
}

// Prints out a string of text graphically using text_img which was loaded from bmp
int TextPrinter::PrintText (const char * str, float x, float y, float font_size, bool centered, int draw_mode, ColorX & color, float angle, float scaley, bool rotate_baseline, bool fixed)
{
	if(font_size <= 0.0f) return 0;

	if(text_img == NULL)
	{
		z_InitWithDefaultImage();
		assertNoLib(text_img);
	}

	float cosAngle = cosf(angle*DEGtoRAD);
	float sinAngle = sinf(angle*DEGtoRAD);

	if(angle == 0.0f)
		rotate_baseline = false;

	if(centered)
	{
		float xp = -GetVisibleStringWidth(str,font_size)/2.0f;
		float yp = -font_size/2.0f;
		if(/*angle != 0.0f &&*/ rotate_baseline)
		{
			x += xp * cosAngle + yp * sinAngle;
			y += xp * sinAngle + yp * cosAngle;
		} else
		{
			x += xp;
			y += yp;
		}
	}
/*
	if(draw_mode & DRAWOP_CENTERED)
	{
		float xp = font_size/2.0f - GetVisibleStringWidth(str,font_size)/2.0f;
		if(angle != 0.0f && rotate_baseline)
		{
			x += xp * cosAngle;
			y += xp * sinAngle;
		} else
		{
			x += xp;
		}
	}
*/
	int numPrinted = 0;

	z_inHighlight = false;

	float xi = x;
	float yi = y;
	float scale = font_size / z_fCharPixels;
	scaley *= scale;
	int n = 0;
	while(str[n])
	{
		int xs, ys; // x and y of source position (top-left)
		float kern;
		bool visible; // becomes false if this character is invisible or unknown and thus we don't have to draw it

		GetVisibleCharInfo(str,n,xs,ys,kern,visible);

		if(str[n] == z_highlightEndChar)
			z_inHighlight = false;

		// offset letter g down a bit, since unfortunately it's based lower than all other characters
///		if(str[n] == 'g')
///			y += font_size/16.0f;

		float xd, yd;
		if(rotate_baseline)
		{
			x -= xi;
			y -= yi;
			xd = x * cosAngle - y * sinAngle;
			yd = x * sinAngle + y * cosAngle;
			xd += xi;
			yd += yi;
			x += xi;
			y += yi;
		} else
		{
			xd = x;
			yd = y;
		}

		if(visible)
		{
			GameX.SetDrawAngle(-angle);
			GameX.SetDrawScale(scale, scaley);
			GameX.SetDrawShading(z_inHighlight ? z_highlightColor : color);
			GameX.SetDrawPart(xs,ys,xs+z_charPixels,ys+z_charPixels);
			GameX.SetDrawMode(draw_mode);
			GameX.DrawImage(text_img, xd, yd);
			numPrinted++;
		}

		if(fixed)
			x += font_size;
		else
			x += font_size*kern;

///		if(str[n] == 'g')
///			y -= font_size/16.0f; // reverse earlier vertical offset for letter g

		if(str[n] == z_highlightStartChar)
			z_inHighlight = true;

		n++;
	}

	return numPrinted;
}


// Prints out a string of text graphically using text_img which was loaded from bmp
int TextPrinter::PrintText3D (const char * str, const Vec3f & pos, const Vec3f & right, const Vec3f & down, float scale, float rOffset, bool centered, int draw_mode, ColorX & color, float scaley, bool fixed)
{
	if(scale <= 0.0f) return 0;

	if(text_img == NULL)
	{
		z_InitWithDefaultImage();
		assertNoLib(text_img);
	}

	scaley *= scale;

	Vec3f posi = pos;

	posi += right * rOffset;

	if(centered)
	{
		float xp = -GetVisibleStringWidth(str,scale)/2.0f;
		float yp = -scaley/2.0f;
		posi += (right * xp) + (down * yp);
	}

	int numPrinted = 0;
	z_inHighlight = false;

	Vec3f xIncr = right * scale;
	Vec3f yIncr = down * scaley;

	bool cull = draw_mode & DRAWOP_NOCULL ? false : true;
	CameraX* cam;
	float leeway; // how far offscreen the top-left corner of the character can go while some part of the character may still be visible
	if(cull)
	{
		cam = GameX.GetDrawCam();

		float x,y,z, x2,y2;
		Vec3f vec = posi;
		cam->Project(vec, x,y,z);
		vec += xIncr;
		vec += yIncr;
		cam->Project(vec, x2,y2,z);

		leeway = max(fabsf(x2-x),fabsf(y2-y));
	}

	GameX.SetDrawMode(draw_mode | DRAWOP_KEEPSTATES);

	int n = 0;
	while(str[n])
	{
		int xs, ys; // x and y of source position (top-left)
		float kern;
		bool visible; // becomes false if this character is invisible or unknown and thus we don't have to draw it

		GetVisibleCharInfo(str,n,xs,ys,kern,visible);

		if(str[n] == z_highlightEndChar)
			z_inHighlight = false;

		if(visible)
		{
			// don't draw offscreen characters
			if(cull)
			{
				float x,y,z;
				cam->Project(posi, x,y,z);
				if(z <= 0 || x < 0-leeway || y < 0-leeway || x >= GameX.GetWidth()+leeway || y >= GameX.GetHeight()+leeway)
					visible = false;
			}

			if(visible)
			{
				GameX.SetDrawShading(z_inHighlight ? z_highlightColor : color);
				GameX.SetDrawPart(xs,ys,xs+z_charPixels,ys+z_charPixels);

				Vec3f ul = posi;
				Vec3f ur = posi + xIncr;
				Vec3f bl = posi + yIncr;
				Vec3f br = posi + xIncr + yIncr;

				GameX.DrawTexturedPolygon3D(text_img, ul, ur, bl, br);

//				GameX.DrawTexturedPolygon3D(text_img, posi, posi + xIncr, posi + yIncr, posi + xIncr + yIncr);

				numPrinted++;
			}
		}

		if(fixed)
			posi += right * scale;
		else
			posi += right * (scale * kern);

		if(str[n] == z_highlightStartChar)
			z_inHighlight = true;

		n++;
	}

	GameX.ResetDrawStates();

	return numPrinted;
}



float TextPrinter::GetVisibleCharWidth (const char * str, int index, float font_size)
{
	return GetVisibleStringWidth(str,index,index,font_size);
}

float TextPrinter::GetVisibleStringWidth (const char * str, float font_size)
{
	return GetVisibleStringWidth(str,0,32767,font_size);
}

float TextPrinter::GetVisibleStringWidth (const char * str, int first_index, int last_index, float font_size)
{
	if(font_size <= 0.0f) return 0.0f;
	int xs,ys; float kern; bool visible;
	float width = 0.0f;
	for(int n = first_index ; n <= last_index && str[n] ; n++)
	{
		GetVisibleCharInfo(str,n,xs,ys,kern,visible);
		width += kern*font_size;
	}
	return width;
}

void TextPrinter::GetVisibleCharInfo (const char * str, int n, int & xs, int & ys, float & kern, bool & visible)
{
	unsigned char c = str[n], nc = str[n+1];
	xs = z_chars[c].xs;
	ys = z_chars[c].ys;
	visible = z_chars[c].visible;

	assert(xs >= 0 && ys >= 0 || !visible);

	kern = (float)z_charKern[c][nc] * (1.0f / 256.0f);
}

void TextPrinter::z_InitWithDefaultImage ()
{
	// hard-code the default font so we can't possibly lose it
	const int width = 512;
	const int height = 192;
	const char* alphabetBits = // vertical run-length encoding of image/8 starting at ';' with numbers signifying repeat of previous
		";472?L;119=C?;13A;51=GTZ2;17CI4GA;28?EIGA;21CTLE;35CRZ2I;13RG;48CNXZ4;14GPZ8XNC;23GVZ5VI;19CZ4RIC;30EZ4I;13RZG;44?IVZ6V;12ATZ13P=;19ATZ9TA;3IRVI;10CZ7XPIA"
		";26VZ3TG;13RZ2I;41EPZ7PE;12GZ16XA;17=XZ11T=;2NZ2P;10CZ11XNG?;21CZ2XA;15NZ3N;37?LVZ6VI=;13EZ18X?;16RZ13I;2EZ3;10?LTZ13VN;19CZ2R;17IZ3;34=GRZ9;15=XZ6TNI3PVZ6R;15?Z5V"
		"NIRZ5T;3Z3C;12=ENVZ11;19=Z2X=;17AVZ;32ANXZ6RLZ3;15IZ4XI=;7ANZ5C;14GZ3VC;4?TZ4;3TZ2G;16?GNVZ7;20PZ2I;18=N;29=IVZ6XL?;CZ3;15XZ3N=;11?TZ3P;14IZ2X?;6=XZ3;3R"
		"Z2I;20ANXZ4;20CZ3?;47AZ7PE;4CZ3;14EZ3N;14=TZ3=;13IZ2L;8RZ2V;3RZ2I;17=ENVZ6;21RZ2N;47CZ4VI?;6CZ3;14IZ2X=;15CZ3C;13EZ2I;8TZ2L;3VZ2I;14AITZ8TI;21CZ3?;46CZ5PG;6CZ3;14"
		"PZ2P;17XZ2I;13=XZP;7CZ2X=;2?Z3G;10?GPXZ7VNE=;24NZ2N;46CZ7XNA;3CZ3;14RZ2I;17RZ2I;14IZ2C;5=TZXA;3RZ3A;9CZ8PIA;28?Z2X;46CZ10TI=CZ3;14RZ2I;17RZ2I;15RZXEC3EV"
		"ZVEC2IVZ3V;10CZ5TG;32=Z3;46=ITZ11TZ3;14RZ2N;17VZ2L;13AXZ19I;10CZ7RIC;27AGTZ2V;49ANXZ12?;13IZ2V;16?Z3X=;12CZ19X=;10CZ10XPIA;23RZ4I;52GPZ11TG=;10GZ3I;15PZ4N;12CZ18XA;12"
		"CITZ11VNG;20RZ3L;55?IVZ11NC;8=Z4G;13NZ6?;11CZ17N?;16=ELVZ10;20LRIA;59CNXZ10;9NZ4RA;9EVZ4XZ2L;11?I14C2;23?GNXZ6;85=GTZ8;9?Z6TLGC3INVZ5XGZ2X;54?LXZ4;88?LVZ5;10"
		"IZ19C;Z3E;50AITZ7;91ENZ3;11NZ17G;2PZ2N;46=ENVZ7TLA;93=IT;12IZ14XC;3IZ2X;43?IPZ8PI?;112?NXZ9VI=;4CZ3A;41CZ7XNE=;118CIPR3NIA;8Z3G;41CZ4TLA;139VZN;42CZPI?;142"
		"RN;43=2;145E;646=C6=;11=C4;168CZ10TR3I2;2CZ4;40CZ23;78?=;14=P;8CZ16;2CZ4;15=EIPR3NIC;15CZ23;8CZ23;7AC24;14CXI;12CVZ;8CZ16;2CZ4;13GTZ11PC;12CZ23;8CZ23;7RZ24;14CZ2RA;9IZ3;8CZ6TR3LI3C2"
		";2CZ4;11CVZ15R?;10CZ23;8CZ23;7RZ24;14CZ4I;6ATZ3X;42GZ18XA;9CZ3R6VZ2XR6TZ3;8CZ23;7RZ24;14CZ5VC;3IXZ3T?;41CZ20X=;8CZ3;6IZ2R;6CZ3;8CZ3I7VZ2RI9;7RZ24;14CZ7N?PZ4I;43VZ4VNIG"
		"C5GIPXZ4N;8CZ3;6IZ2R;6CZ3;8CZ3;7RZ2I;16GI9LZ2XLI10;14=NZ11XC;43EZ3TC;13GXZ3=;7CZ3;6PZ3;6CZ3;8CZ3;7RZ2I;25AXZT?;27CVZ8T=;44LZ2T;16AXZ2E;7CZ3E;4?Z4G;5IZ3;8CZ3;7"
		"XZ2VE;23AXZX=;30IZ7XI;44RZ2E;17LZ2I;7CZ3T=;2ATZ4XA;4VZ3;8CZ3A;5CZ5P=;21RZ2I;32LZ8TC;42RZ2C;17IZ2I;8Z4XRTZ3NRZ2XICERZ3V;8CZ3I;4=TZ6XG;19AZ3C;30?RZ11L=;40"
		"PZ2I;17RZ2I;8TZ9CGZ10P;8AZ4I;2ETZ9R?;17IZ3I;29GXZ3XGPZ7X;40IZ3E;15LZ3A;8GZ8L;=VZ9G;9Z11XIZ6XI;16IZ4NI12;14=PZ4P=;2EVZ6;40?Z4RG?;9AIVZ3V;10NZ6P=;2EZ8T;10PZ10"
		"I;AVZ6TA;14GZ17;14CZ3XE;5=LZ5;41LZ7TR5XZ7E;11ETZ2RG;5EXZ5R=;10AXZ8N;3=LZ7;14=XZ16;14CZ2P=;8ATZ3;41=TZ19N;24=IR2NE;13AXZ5XI;6EXZ5;15IZ16;14CXE;12IXZ;42=RZ17L;45="
		"IPRPI=;8=RZ4;16GVZ14;14?=;14?P;44GVZ13RC;63IXZ2;18?C13;78ENTZ7RIA;66ATZ;114AC3?;72I;2284?I3A;10AI3C;105IZR;62=NZ5RA;8=Z3R;8C3A;2=C17;73PZR;36=CI4GC;17=TZ7XA;8NZ3;8Z3"
		"R;2CZ17;14=?;34CLRZ3TNC;14RZR;34GRZ8XNC;14NZ9V=;7EZ3A;7Z3R;2CZ17;14CZTG=;29GVZ9RA;12TZ22;12CVZ13R=;11?Z11P;8Z3C;7Z3R;2CZ17;14CZ4PG=;24ATZ2VLC=;2CLXZXG;11Z23;11G"
		"Z16XA;10IZ3TE2RZ5C;7TZ2I;7Z3R;2CZ17;14CZ7NE;12=L;7GZ2VG;8=NZ2A;10Z23;10GZ18XA;9NZ2X=;3TZ4V;7RZ2I;7I3G;2?I17;14CZ9XNC;8GXZ;6NZ2I;12NZV;9?Z23;9?XZ6TNI3PVZ6T;9RZ2"
		"N;4?Z5G;6RZ2I;45AVZ11VLA;3?RZ3;5GZ2G;6CI3?;3VZG;8AR23;9NZ4XI=;7ANZ5G;8RZ2I;5IZ4X=;5XZ2I;47ALVZ11VI2Z5;4AZ2G;4?NXZ5G;2GZP;41XZ3P=;11CXZ3T;8RZ2I;5=XZ4N"
		";4CZ3E;50CNXZ15X;4TZL;4IZ8X;3Z2;40EZ3T;14CZ4=;7LZ2P;6IZ5N=;GXZ3?;53ENXZ10VI=;3GZV=;2=RZ3TNILXZ2;3TZA;39IZ3?;15NZ3C;7IZ2X;7TZ11V;57GXZ6VG;6RZG;3RZXI"
		"?;4EZP;3RZC;39PZ2T;16AZ3I;7CZ3E;6AZ11G;55=IVZ5RG;7?ZX;3IZT?;6RZC;3RZC;39RZ2L;17XZ2I;8Z3N;7IZ9T;54=IXZ5RC;9IZN;3VZ?;5=NZI;4XZ?;39RZ2I;17RZ2I;8GI3;8"
		"LZ7R=;52?NZ6NA;11LZI;2AZR;5?TZI;4?ZV;40RZ2I;17RZ2I;21CNXZ2PG;52CNZ6N?;13RZC;2CZT;4IXVE2IA;2IZL;40RZ2I;17RZ2I;78CRZ5XI?;15RZC;3Z2?CIRZ2PZ3T;2PZ?"
		";40LZ2R;17XZ2C;77CZ5XI=;17RZC;3RZ12;3=;41IZ2X;16?Z3?;77CZ3VG;20LZI;3RZ6XRITZ2;45EZ3A;15IZ2X;78CZTG;22GZT;3RZXRIE?;3CZX;45?Z3I;15VZ2P;78?C;24=Z2A;2A?;8GZP;46"
		"VZ2T;14GZ3E;105NZR;12RZG;64AC3;106=XZN;10NZT;176CZ2P?;6?PZXA;177GXZ2NGC2GNZ2XA;179?RZ8N=;182AIPR2NG=;1190AC3;8CZ23;137IR2I;14IZ4;8CZ23;8AR3;53?C2A;24C3;23LTC;15VZ2C;13NZ5"
		";8CZ23;8CZ3;53?Z2X;9AR2L;9=TZ3;16CE;5Z2C;14AZ2X;12=RZ6;8CZ23;8CZ3;54TZ2C;8CZ2R;8ATZ4;15IZR;5Z2C;14GZ2N;12PZ7;8CZ3R16TZ3;8CZ3;54LZ2I;8CZ2R;7AXZ5;15IZR;5Z2C;2CIRZ;8IZ2"
		"I;11NZ4XZ3;8CZ3;16CZ3;8CZ3;54IZ2I;8CZ2R;6GXZ6;15IZR;5Z2NRZ5;8PZ2C;10IZ3TAIZ3;8CZ3;16CZ3;8CZ3;54TZ2I;8CZ2R;5GZ8;15IZR;=CIRZ6TLE;8RZ2C;9GZ3T=;IZ3;8CZ3;16CZ3;8C"
		"Z3;28C3A;2=C17GRZ3G;8CZ2R;4NZ5TZ3;15IZXTZ6TLE=;11RZ2E;8GZ3T=;2IZ3;8CZ3;16CZ3;8CZ3;28Z3R;2CZ22A;8CZ2R;2=NZ5P=Z3;11=ELTZ6TLZ2C;14RZ2N;7GZ3X=;3IZ3;8CZ3;16CZ3;8CZ3"
		"I20;8Z3R;2CZ21R;9CZ2R;=TZ5N;2Z3;8?LTZ6XIC;3Z2C;14LZ3A;4=LZ4C;4IZ3;8CZ3C;15IZ3;8CZ23;8Z3R;2CZ20XA;9CZ2RATZ5I;3Z3;8CZ4RINZR;5Z2C;3CIR;8GZ3XIC2LVZ4I;5IZ3;8C"
		"Z3I;15NZ2X;8CZ23;8Z3R;2CZ19VA;10CZ2VXZ5G;4Z3;8CRIC;3IZR;5Z2I2RZ4;8=Z12N;6IZ3;8=Z3V;14=XZ2R;8CZ23;8I3G;2?I17GA;12CZ7XA;5Z3;15IZR;3CIZ7TL;9LZ10T=;6IZ3;9XZ3N;13PZ3L"
		";8CZ23;46CZ6VA;6Z3;15IZVLTZ6TLE=;11=TZ8T=;7IZ3;9PZ4P=;9?RZ4A;8CZ3C20;46CZ5T=;7Z3;12=ELVZ6TZ2C;16?TZ6L=;8IZ3;9CZ5XNE?;3CGPZ5R;9CZ3;66CZ4P=;8Z3;8=ELTZ6RI"
		"C=;Z2C;18GNR2I?;10CI3;10RZ18XA;9CZ3;66CZ3N;10Z3;8CZ5R2ZR;5Z2C;48=XZ17G;10CZ3;66=C2A;11C3;8CZRIC;2IZR;5Z2C;49?TZ14XC;11CZ3;92=;6IZR;5Z2C;51IXZ10XI=;12CZ3;99IZR;5IC="
		";52=GNTZ4RNG=;14CZ3;99ITI;86CZ3;188?I3;1755AC24;136CZ16TPG=;10RZ24;22G?;35?C2A;29?C2A;40CZ19XC;9RZ24;22VN;20GNRPI=;9?Z3;10VZ2?;15EZ2X;8CZ23;8CZ21E;8RZ24;21GZ2?;17?TZ5XE;9NZ2E"
		";8AZ2T;16=Z3?;7CZ23;8CZ21X=;7RZ24;21VZ2N;16=VZ8E;8EZ2L;8IZ2L;6RZ2C;7TZ2C;7CZ23;8AR16Z6L;7GI14N2I8;20GZ4?;15IZ9XA;8XZR;8IZ2I;6RZ2C;7PZ2I;7CZ23;26APZ3X;21AT2A;27VZ4N;15X"
		"Z2XGCLZ4T;8RZX;8PZ2C;6RZ2C;7IZ2I;7CZ3R6VZ2XR6TZ3;28NZ3C;19IXZ3I;25GZ2XZ3?;13AZ3A;3IZ4I;7IZ2;8RZ2C;6VZ2E;7IZ2I;7CZ3;6IZ2R;6CZ3;28=Z3G;17=PZ6TA;23VZ2IRZ2N;11R"
		"Z4XI5XZ16I;5RZ2C;5=Z3L;7LZ2I;7CZ3;6IZ2R;6CZ3;29TZ2I;16EXZ9I;21EZ2X=CZ3=;10RZ27I;5RZ2L;5IZ3V;7VZ2I;7CZ3;6IZ2R;6CZ3;29RZ2I;14=LZ4LTZ6TA;19RZ2I;2RZ2I;10GINZ2RI6L"
		"Z4P;4=XZ2NIC;5PZ2XA;3CXZ4I;5IZ3E;7CZ3;6IZ2R;6CZ3;29VZ2I;13?TZ3VA;=LZ7I;17CZ2X=;2CZ2X=;11CZ2L;7PZ4N;2=NZ3=;7IZ3XNIPZ3TZ3NA;=IZ4A;7CZ3;6IZ2R;6CZ3;28C"
		"Z3G;13CZ3N=;4EXZ6;17RZ2I;4RZ2I;12Z2R;7=XZ4XRXZ3P;8AZ9XAZ11V;8CZ3;6IZ2R;6CZ3;27?TZ3A;13CZXE;7=PZ5;16CZ2X=;4CZ2X=;11VZ2;8CZ9X?;9VZ8I;RZ10I;8CZ3;6IZ2R;6CZ3;8="
		"C16INXZ3V;14CR=;10IXZ3;16RZ2I;6RZ2I;11NZ2E;8GZ7XG;10CZ7T;2?XZ8T;9CZ3;6GR2L;6CZ3;8CZ22G;14=;13?TZ2;15CZ2X=;6CZ2X=;10CR2I;9EVZ4N?;12EXZ4R=;3AXZ6R=;9CZ3;16CZ3;8C"
		"Z21R;31IZ;15RZ2I;8RZ2I;25?C2;16=GI2A;6=LVZ2PG;11?I3;16CZ3;8CZ20R=;32A;14CZ2X=;8CZ2X=;92=C3;8CZ18RG;171=C16?;2509AC24;104CZ23;39RZ24;43?IR3I?;35GZ3R;14CZ23;8AI?;28RZ24;14CZ3=;8CZ2X"
		"=;10IZ7I;33NZ4R;14CZ23;8CZ2PE;26RZ24;15RZ2I;8RZ2I;10IZ9I;11=A;17=TZ5R;14CZ23;8CZ4XL?;23RZ24;15CZ2X=;6CZ2X=;10VZ3R3Z3V;10ATI;16AXZ3XZ2R;14CZ3R7Z3TR9;8CZ7TG=;20GI24;16RZ2I;6"
		"RZ2I;10AZ2N=;3=NZ2A;8IXZI;15GZ3XGRZ2R;14CZ3;7Z3C;17CZ10NC;59CZ2X=;4CZ2X=;10CZ2;7Z2C;6=PZ2TA;13=NZ3VA;RZ2R;14CZ3;7Z3C;17?NXZ10VI?;57RZ2I;4RZ2I;11CZ2G;5GZ2C;5"
		"AVZ2L=;13?TZ3R=;2RZ2R;14CZ3;7Z3C;19=GTZ11PG;55CZ2X=;2CZ2X=;12XZ2TI3TZ2X;5IZ2XE;14CXZ3L;4RZ2R;14CZ3;7Z3C;22?LVZ10XNA;53RZ2I;2RZ2I;13LZ9L;3=RZ2T?;14IZ3XG;5RZ2"
		"R;14CZ3;7Z3C;25EPZ11TI=;50EZ2X=CZ3=;13=RZ7R=;2CXZ2I;14=NZ3VA;6RZ2R;14CZ3;7Z3C;27=IVZ11;51VZ2IRZ2N;16GRZ3RG;3LZ2XC;14=TZ3R=;7RZ2R;14CZ3;7Z3C;30ANXZ8;51GZ2XZ3?;24"
		"?TZ2P=;15CZ4R9XZ9;8CZ3;7R3A;33GRZ6;52VZ4N;24GXZ2I;17CZ23;8CZ3;44?LVZ5;52GZ4?;22=NZ2VA;2=C2A;12CZ23;8?I3;41=ITZ7R;53VZ2N;22ATZ2N=;ARZ4XI;10CZ23;51GPZ7VI?;54GZ2?;21IZ2"
		"XG;2EZ8T;23RZ2R;54CNXZ6XLA;58VN;20=PZ2TA;3XZ9I;22RZ2R;51?LVZ7NE;61G?;19CVZ2L;4EZ2VIC2ENZ2T;22LR2L;48=ITZ7RG=;83IZ2XE;5IZV=;5IZ2;72?PZ7VI?;84?RZ2R=;6I"
		"ZT;6EZ2;72CZ5XLA;87RZ2I;8GZ2PA;2=GXZV;72CZ3NE;90RVC;9=Z10L;72CRG=;92I=;11IZ8X=;182IXZ5R?;184=EI3A;138=;959=;49?;78ALGA;59RN?;45CR2;34CI2G;40CZ3XRIE=;29=C17;7RZ2P"
		"C;41CRZ2R;14CZ11I;7AZ3;16=CI4GC;16CZ8VNIC;25CZ17;7RZ4RC;37CRZ4R;14CZ11E;8XZ2A;13GPZ8XNC;13CZ13RLG?;20CZ17;7NZ6VG;33CRZ5N?;14CZ11G;8RZ2G;11CVZ13P=;11CZ17XRIC=;15CZ17"
		";8CNZ6VG;29CRZ5N?;16CZ3NI2C2NZ2I;8IZ2I;10GZ16XA;11?GLRZ18V;14CZ17;10?NZ6XI=;24CRZ5N?;18CZ3C;4IZ2L;8IZ2I;9GZ18XA;15CINVZ14;14?I2NZ2VI11;12?IXZ5XI?;20CRZ5N?;20CZ3"
		"C;4IZ2V;8LZ2I;8?XZ6TNI3PVZ6T;19=CIRXZ9;16GXZT=;25=IXZ6N?;16CRZ5N?;22CZ3C;4CZ3C;7TZ2I;8NZ4XI?;7APZ5G;23?ITZ6;15EZ2V=;29GVZ6NC;13=IXZ4VG;23CZ3C;4=Z3V"
		"=;5GZ3C;7=XZ3T=;11EXZ3P;20AINVZ8;15XZ2G;32GVZ6RC;13=IXZ4VG;21CZ3C;5TZ3VG=;ANZ4=;7EZ3T=;13EZ4;16CIRXZ9TLE;14EZ3C;34CRZ6TG;13=IXZ4VG;19CZ3C;5IZ12P;8IZ3C"
		";15NZ3C;10?GLTZ9XPIC;18IZ3P;36CPZ6VG;13=IXZ4VG;17CZ3C;5=VZ11C;8NZ2V;16AZ3G;7?PVZ9VNGA;22IZ4XR12;25?NZ6XI=;12=IXZ4VG;15CZ3C;6AXZ9L;9RZ2N;17Z3I;7CZ7RIE=;26"
		"GZ17;27?LXZ5XE;13=IXZ4P;14CZ3C;7AXZ7I;10RZ2I;17RZ2I;7CZ8TNIC;25=XZ16;29=IXZ4I;15=IXZ2R;14?I3?;8=ITZ2VNA;11RZ2I;17RZ2I;7CZ13RLGA;21GZ16;31=GVZ2I;17=IXR;46RZ2I"
		";17RZ2I;7ARZ17RLGA;17ETZ14;34GVI;19=E;46NZ2P;9CI7Z3G;10CINTZ16X;17IZ2I;47=;67IZ2V;9IZ10C;15CINTZ12;16NZ2G;116EZ3?;8IZ10=;19=CIPVZ7;15LZ2N;117AZ3I;8IZ9V;23AITZ6;14?Z3C;118"
		"XZ2P;8IZ9P;19=GNXZ8V;14GZ3G;118AC3;8?C9?;16AIRZ9PI?;15IZ3XIC12;140=ENVZ8VNE=;18IZ17;137?IPZ9TIA;22CZ17;136CZ8XNG=;26TZ16;136CZ5TLC;30=TZ15;136CZ2PG?;35AI14;136?E;996=C17;79A"
		"IRTZ4TPI?;13CZ23;31?;14CZ17;7LR28G;16=ITZ2TI=;16GVZ12N?;11CZ23;8CL=;19IX;14CZ17;7RZ28I;15GXZ6XE;13ATZ16I;10CZ23;8CZXG;16?RZ2;14CZ17;7RZ28I;14EZ10E;11AXZ18N;9CZ23;8CZ3RA;13I"
		"XZ3;14CZ17;7AC28?;13=XZ10X=;9?XZ6VRVZ11E;8AR9XZ2VR10;8CZ5I=;9?RZ4T;14?I2LZ2XLI10;42=NXZTI=;IZ3VI2PZ5I;9RZ4RG=;GZ2RE;2?GVZ3T;18RZ2I;18CZ6VE;7IXZ4I;17AXZ"
		"T?;52AXZ5XIXZ2N;4=PZ3V;8?Z3XE;3?XZP;7NZ3?;17RZ2I;18?VZ7R?;3ATZ4VA;17AXZX=;53TZ10R;7PZ3=;7IZ3C;4NZ2=;8TZ2G;17RZ2I;19=IZ7XI;IZ5L=;18RZ2I;53EZ11N=;6=Z3C"
		";7NZ2N;5Z2R;9IZ2I;17RZ2I;21?RZ12VC;19AZ3C;53LZ3NIPZ6XI;6RZ2C;7RZ2G;4CZ2R;9IZ2I;17RZ2I;23GXZ9N=;20IZ3I;53RZ2I;3?RZ6VC;4RZ2C;7RZ2C;4CZ3?;8TZ2I;17RZ2I;24=LZ7XI"
		";21IZ4NI12;40RZ2C;3=TZ8P=;2VZ2=;7RZ2C;4AZ3T=;6LZ3C;17RZ2I;25?TZ8TC;19GZ17;40NZ2P;2=PZ2LNZ6VC?Z2V;8RZ2G;5XZ3XLEC2IRZ3X;18RZ2I;24IXZ11N=;17=XZ16;40GZ3VRZ3X=;C"
		"VZ6TZ2L;8IZ2I;5NZ13L;18RZ2I;22?RZ4XGNZ7XI;17IZ16;40=VZ7E;3=LZ8?;8GZ2R;5?XZ11T=;8?I9VZ2RI10;11GXZ4P=;2CTZ7RA;16GVZ14;41AXZ5G;6EXZ5XC;8CZ2X;6GZ10TA;9CZ23;9=PZ4"
		"VE;6IXZ7;18?C13;42APZ2TC;8?TZ6;9I3?;6EVZ7L=;10CZ23;8?XZ4L=;8?PZ6;88CTZ6;21GNR2NI=;12CZ23;8CZ3VC;12CVZ4;83C2GITZ3X2Z3;40CZ23;8CZ2L=;15IXZ2;83Z7XA2TZ2;40=C23;8CVC;18?R"
		"Z;83Z6P=;2=PZ;72=2;21E;83Z4NC;6N;179I2C;1554?I3?;59=;127CZ3C;18C;8CZ23;8CTC;41?EI3C=;15GNR5;55CZ3C;14=ITZ2;8CZ23;8CZ2N=;37ETZ7PA;11=TZ7;55CZ3C;12=LXZ4;8CZ23;8CZ3XI;34=PZ11I"
		";10TZ8;55CZ3C;11IXZ6;8CZ23;8CZ5TC;32PZ13I;8GZXICTZ4;27=GI;25CZ3C;9ATZ8;8AR23;8CZ7P?;29GZ15?;7PZC;2RZ4;26GZ2T;3A;21CZ3C;7=LZ8TR;40=PZ7XI;28VZ4VPI3RXZ4N;44VZ2;2GZ"
		"C;20CZ3C;6CVZ5XLC;45CVZ7VC;25AZ3VE;7IXZ2X;44GZ2=CZ2X=;19CZ3C;4=LZ5RG;50IXZ7RC10;14GZ2X?;9CZ3?;9GNR5;24C=;2RTGXZ2RA;19CZ3C;3CVZ4RC;53=NZ17;14IZ2N;11VZ2C;7="
		"TZ7;24Z2XRE;2TRC;21CZ3C;=PZ4RC;57CTZ15;14IZ2I;11RZ2C;7TZ8;24Z4C;2C=;22CZ3CEXZ3VG;61IXZ13;14IZ2P;11XZ2C;6GZXICTZ4;24ZRNIE=AX2I=;20CZ3TZ4L=;62IXZ13;14EZ3C;9IZ3"
		"=;6PZC;2RZ4;27=VZETZ2XC;19CZ6VA;62CTZ4VI10;14?Z4L?;5APZ3V;44IZ2;?XZT;20CZ5L=;61?PZ5I=;25RZ5XR3Z6I;43=XZX;2AV?;20CZ3XE;62IXZ4R?;27CZ14X=;43ETZR;3=;21=C3?"
		";61CVZ4VE;30IZ13C;46?E;89=PZ4XI;33GXZ9VA;138CZ4P?;35?LVZ5TI=;139CZ2VC;40=C3;143CXI;189?2;2148=C23=;45C4;32AI2E=;81CZ23C;41GNVZ6VNE;16?I2G=;5ATZ4XG;34CZ3G;23EI3;14CZ23CLR4;3"
		"PR;28=IXZ12XI=;11=PZ4XE;3GZ8G;34VZ2I;3AR3;14=PZ4;14CZ23CRZ4;?NZP;27IXZ16XI;9=TZ7G;CZ10C;33NZ2P;3CZ3;13CVZ5;14CZ23CRZ8C;25?TZ20T?;7NZ9CVZ3TRVZ4R;33IZ2R;3CZ3;11"
		"=LZ7;14?I2NZ2PI6TZ2LI7?RZ7G;25AXZ5TNIEC4EINTZ5XA;5?Z4RVZ4XZ2T?;3CXZ3=;32LZ2R;3CZ3;10AVZ8;16GXZN;7=TZT=;7RZ4XPC;25AXZ3TG=;12=GTZ3XA;4IZ2XA;2IZ5T"
		"=;5EZ3C;31=VZ2R;3CZ3;9IZ10;15EZ2V;9EZ2I;40TZ2VE;18EVZ2T;4NZ2I;4LZ4E;7VZ2I;7=C22IVZ3I;3CZ3;7ATZ6XNZ3;15XZ2G;10Z2X;39IZ2R=;20=RZ2I;3RZ2C;4=VZ3R;7LZ2I;7CZ27C;3CZ3"
		";6IZ7R?CZ3;14EZ3C;10Z3C;6LR4;3PR;55RZ2C;5GZ4C;6IZ2I;7CZ26R;4CZ3;4?RZ7I;2CZ3;14IZ3I;9GZ3C;6RZ4;?NZP;55RZ2I;5CZ4T;6NZ2I;7CZ25XA;4CZ3;3IXZ6TA;3CZ3;14IZ3XA;7EX"
		"Z3A;6RZ8C;55LZ2XA;3EXZ5I;4=XZ2C;7CZ24R?;5CZ3;=PZ7I;5CZ3;14GZ5NIC2EITZ4X;7RZ7G;56GZ3XLITZ7XA;3NZ3=;7?I22C?;7CZ3EXZ6VC;6CZ3;14?Z16L;7RZ4XPC;57=XZ8XIZ5N"
		"IRZ3P;40CZ10L=;7CZ3;15LZ14V=;73IZ7XA;RZ10A;40CZ8XC;9CZ3;15=RZ12T?;75NZ5VA;2=TZ8G;41CZ7P=;10CZ3;17IXZ8XI;78CNRPG;5=RZ5XG;42CZ5XE;12CZ3;19EIR3PIE;92GNR2I=;43CZ4"
		"R=;13CZ3;168AR3I;15AR3;2516?CI4C;44IR2I;22CI2E;10C3;22CZ23;25AI2G=;20GTZ7XN?;41=XZ2P?;18=IZ2X?;7CPZ5PC;19CZ23;23=PZ4XG;17?RZ11XE;10=INR5A;22CZ3XL?;14=IVZ3G;6=NZ9I;6G"
		"Z2V;8CZ23;15IZ2R;4TZ7E;15=XZ14A;8AXZ7C;23GZ5TLGC=;4=CGITZ5G;7TZ11I;5=Z3;8CZ23;15TZ2C;3IZ8V;15PZ15R;7=XZ8C;24EXZ21G;7LZ13?;5TZ2C;7AR10TXR11;14=Z2V;4VZ9A;13?Z5XPI3R"
		"Z6=;6IZ2NIVZ4C;25?RZ18RA;7?Z5VR2VZ5L;5PZ2G;17?TZL;24CZ2P;3AZ3TE2TZ3C;13GZ3XG;6=NZ4C;6RZC;2RZ4C;27CRZ14VG;9IZ3P?;4?PZ3T;5IZ2I;16EXZ3T?;22IZ2I;3EZ2X?;2?"
		"Z3A;13IZ3C;9TZ3C;6AC;3AC4=;29ANVZ8VNC;11NZ2R;8NZ3;5IZ2I;15LZ6XE;21IZ2I;3IZ2P;4Z2X;14IZ2T;10IZ3?;49?CI4C?;14RZ2E;8?Z3;5LZ2I;13=TZ9L;20IZ2N;3IZ2I;3CZ2I;14EZ2R;10L"
		"Z2V;72RZ2C;9Z2T;5TZ2G;12CXZ3VTZ6R=;18IZ2XA;2IZ2I;3TZR;16XZX;9=XZ2E;72PZ2I;8AZ2L;4EZ3A;11IZ4P=;LZ6XC;17CZ4R2VZ2VR2TZX?;16EZ2I;8NZ2G;73IZ2V=;7PZX?;3EXZ2"
		"V;10=RZ4I;4EXZ6I;16=Z15XA;16LZ2IC6NZXIC8=;65AZ3XG=;3=PZ2G;2ANZ4I;9AVZ3XC;6?TZ6R=;15NZ15V;14AVZ22C;66RZ5TRTZ3RIPVZ5V=;8AZ4R=;8=NZ6X;15=XZ15A;13CZ23C;66AX"
		"Z18XA;9CZ3I;12GXZ5;16?NXZ13C;13CZ23C;67AXZ16T?;10CZXE;14ATZ4;28?XZ2A;13CZ23C;68=NZ13VG;12CT?;16=NZ3;29RZX;14?I23?;70?IRZ7VNE;14=;20GXZ;145?C4;40AT;191=;2311AC24;14=C17;135RZ24;14C"
		"Z17;64R2A;37CZ23;7RZ24;14CZ17;7LR4A;2IRA;46Z2C;17GR2G;16CZ23;7RZ24;14CZ17;7RZ4C;GXZ?;46Z2C;17IZ2I;16CZ23;7RZ23X;14CZ17;7RZ8R;47Z2C;17IZ2I;16CZ23;7GI9NZ2PI6TZ2G;15?I2TZ2TI11;7RZ7XA;47"
		"Z2C;17IZ2I;16AR19TZ3;16GXZN;7=TZT=;16NZ2N=;18RZ6RA;48Z2C;17IZ2I;36CZ3;15EZ2V;9EZ2I;15IZ2V=;19AC5;51Z2C;17IZ2I;36CZ3;15XZ2G;10Z2X;14=XZ2I;77Z2C;17IZ2I;36CZ3;14EZ3C;10Z3C;13EZ3"
		"C;77Z2C;17?C2?;36CZ3;14IZ3I;9GZ3C;13IZ3E;77Z2C;57CZ3;14IZ3XA;7EXZ3A;13GR3I;77Z2C;57CZ3;14GZ5NIC2EITZ4X;96Z2C;57CZ3;14?Z16L;96Z2C;57CZ3;15LZ14V=;96C2=;57CZ3;15=RZ12T?;157CZ3;17IXZ8"
		"XI;159=C3;19EIR3PIE;2778AC2=;188RZ2C;47CZ23;20?EI3C=;22CNRPE;6IR2I;7AC28A;16RZ2C;25RZ2C;2IZ2I;12CZ23;18CRZ7N?;19NZ5N;5IZ2R;7RZ28R;16RZ2C;25RZ2C;2IZ2I;12CZ23;17NZ11G;17IZ7E;4"
		"AZ3;7RZ28R;16RZ2C;25RZ2C;2IZ2I;12CZ5VR17;16NZ13G;16XZ7V;5Z3;7RZ28R;16RZ2C;25RZ2C;2IZ2I;12CZ6TI?;29CZ15?;14CZ9G;4TZ2C;6RZ2TR22TZ2R;16RZ2C;25RZ2C;2IZ2I;12CZ9RG=;26RZ5RI2LT"
		"Z5N;14IZ2XC=PZ3T;4RZ2C;6RZ2C;22CZ2R;9AC6TZ2IC6=;18RZ2C;2IZ2I;12=NXZ10PE;23?Z4P?;5CTZ3X;14IZ2N;2=Z4?;3RZ2C;6RZ2C;22CZ2R;9RZ16C;18RZ2C;2IZ2I;15ENXZ9XNE;20CZ3L"
		";8=VZ3?;13IZ2I;3NZ3N;3XZ2C;6RZ2C;22CZ2R;9RZ16C;18RZ2C;2IZ2I;17=GPZ10VLA;17IZ3=;9GZ3C;13IZ2I;3EZ4C;NZ3=;6AC2=;22=C2A;9LR6XZ2TR6A;18RZ2C;2IZ2I;20=GTZ10VC;15IZ2R"
		";10=Z3C;13GZ2N;4VZ8V;53RZ2C;25RZ2C;2IZ2I;23?ITZ8I;15IZ2R;11Z3C;13CZ2T;4GZ8I;53RZ2C;25RZ2C;2IZ2I;26ALVZ5I;15IZ2T;11Z3?;13?Z3=;4VZ6V=;53RZ2C;25RZ2C;2IZ2I;27ENXZ4I;15CZ3"
		";10AZ3;15C3=;4AXZ4VA;54RZ2C;25RZ2C;2IZ2I;24ENXZ6TC;15=Z3A;9GZ2R;24?LR2I=;55RZ2C;25RZ2C;2IZ2I;21ANXZ6TI?;18NR2E;9IR2I;85RZ2C;25RZ2C;2IZ2I;18ALVZ6TI?;123LR2A;25RZ2"
		"C;2IZ2I;15ALVZ6VL?;155RZ2C;2IZ2I;12=IVZ6VLA;180CZ5VLA;183CZ23;168CZ23;168CZ23;168CZ23;168?I23;1486=C4?;7AC4;3AC;73?I2G;73R5;3LR;9CZ4I;7RZ4;2=X2;73CZ2R;73Z5;2GZV;9CZ4I;7RZ4CGVZ"
		"N;3CZ23;20?CI4C;17I2NZ2VI10E=;9LR2A;22AR2L;22Z5RVZ2I;9CZ4I;7RZ7X=;3CZ23;18GTZ7XN?;13?Z17XG;8RZ2C;22CZ2R;22Z8T=;9CZ4I;7RZ6R?;4CZ23;16?RZ11XE;12CZ19E;7RZ2C;22CZ2R;22Z7N"
		"=;10?I4C;7GI5A;6CZ7TR15;15=XZ14A;11EZ19T;7RZ2IC22IZ2R;22C5A;39?VZ6XE;29PZ15R;11IZ20=;6RZ28R;68=LZ7P=;26?Z5XPI3RZ6=;10?C2IZ2TC9IXZ3C;6RZ28R;70CVZ6XE;25GZ3XG;6=NZ4C;13C"
		"Z2R;10CZ3C;6RZ28R;71=LZ7P=;23IZ3C;9TZ3C;13CZ2R;11Z3A;6LR28L;73CVZ6XE;22IZ2T;10IZ3?;13AR2L;10=Z2X;111=LZ7P=;20EZ2R;10LZ2T;28=C2A;113CVZ6XE;20XZX;9=XZ2C;146=LZ7P=;18EZ2I"
		";8NZ2G;149CVZ6XE;9AC8NZ2IC6NZXIC2;149=LZ7P=;7RZ24;151CVZ6X;7RZ24;136CZ23;7RZ24;136CZ23;7RZ24;136CZ23;7GI24;136AR23;2254=C4?;7AC4;52?GI3C;97R5;14CZ4I;7RZ4;17CI4GA;26GTZ6XN?;17=C13=;23RZ2"
		"I;35Z5;14CZ4I;7RZ4;14GPZ8XNC;21=RZ10XG;16CZ14RA;12C2?;6TZ2L;6ACA;26Z5;14CZ4I;7RZ4;12ATZ13P=;19TZ13G;15CZ15XA;9IXZ3XPIC2NZ4ICEIRZ4VE;24Z5;14CZ4I;7RZ4;11GZ16XA;17IZ14X="
		";14CZ16R;8IZ12XZ13C;23Z5;14?I4C;7GI4;10EZ18X?;16XZ3XPNZ2NPZ5I;14CZ17?;6?Z13ELZ12V;23C5;41=XZ6TNI3PVZ6R;15CZ3N=;CZ2C;=RZ3T;14?I12RZ4C;6IZ12G;2NZ12A;68IZ4XI=;7ANZ5G;14I"
		"Z2R;3CZ2C;2=TZ3;28RZ3A;6NZ2XECGIR3L?;4CLR2PIECIZ3G;68XZ3N=;11?TZ3P;14IZ2I;3CZ2C;3GZ3C;27IZ2X;7RZ2G;22NZ2I;67EZ3N;14=TZ3=;13IZ2N;3CZ2C;3=Z3C;27PZ2I;7AC2=;22"
		"?C2?;67IZ2X=;15CZ3C;13EZ3I=;CZ2C;4Z3C;26CZ2T=;104PZ2P;17XZ2I;13=Z5RTZ2C;4Z3C;25CXZT=;105RZ2I;17RZ2I;14NZ8C;3=Z3=;13?I10NZ2XI3;104RZ2I;17RZ2I;14=TZ7C;3CZ3;14CZ17;104RZ2N;17"
		"VZ2I;15=RZ6C;3IZ2R;14CZ17;104IZ2V;16?Z3E;17AIRTZ2C;3NZ2N;14CZ17;104GZ3I;15PZ3?;45CZ17;104=Z4G;13NZ3T;46=C17;105NZ4RC;9GVZ4I;169?Z6TLGC3INXZ5V=;170IZ19C;172NZ17G;174IXZ13VC;176?NX"
		"Z9VI=;180CIPR3NIA;1614AC2=;22?C2?;31=;35CZ23;71RZ2G;22NZ2I;29CRI;35CZ23;14CZ2R;28?A;23NZ2XECGIR3I?;4CLR2PIECIZ3G;27GVZ2I;4VZ2I;27CZ23;14CZ2R;28CZVI?;20IZ12G;2NZ12A;25"
		"GVZ4I;3AZ3?;27CZ23;10ELR2TZ2XR14;14CZ4RG=;17?Z13ELZ12V;23=IXZ6G;3IZ2P;28CZ3R7TZ3R9;8AVZ22;14CZ6XNC;16IZ12XZ13C;21=IXZ5XI?;4LZ2I;10AI2;2?I4;8CZ3;7CZ3;16=VZ23;14CZ9VL?;14"
		"IXZ3XPIC2NZ3XICEIRZ4VE;20?NZ6XI=;6RZ2C;8ATZ3;2CZ4;8CZ3;7CZ3;16IZ24;14?TZ11TG=;13C2?;6TZ2L;6AC2;20?NZ6VG=;8RZ2C;7EXZ4;2CZ4;8CZ3;7IZ3;16RZ24;16?LVZ11NE;20RZ2I"
		";27CPZ6VG;11RZ2G;6GZ6;2CZ4;8CZ3A;6RZ2X;16Z3TA;2CZ2R;33CNXZ10;49CRZ6RE;13RZ2P;5IZ4VR2;2AR4;8CZ3I;5GZ3R;16Z3?;3CZ2R;35=GRZ8;47GTZ6RC;15IZ3I=;ERZ3TA;18?Z4I?;?I"
		"Z4I;16Z3;4CZ2R;38?IVZ5;45GVZ6NA;17GZ10T=;20Z13?;16XZ2=;3CZ2R;39ANXZ4;42=IVZ6N?;19=Z9T=;21PZ11I;17RZ2E;3?I2G;36=GTZ6P;40=IXZ5XI?;22LZ7T=;22AXZ9R;18NZ2L;41CNXZ5VI="
		";40IXZ5XI=;24=TZ5T=;24EXZ7L;61=IVZ6NC;43RZ4VG=;27=LXZVI=;26=IPR2PI?;60ENZ6TI=;45RZ2VG;130CZ5XNA;48R2E;132CZ3TG=;50?;134CXL?;188=;1975"
	;

	text_img = new ImageX();
	z_ownsImage = true;

	text_img->Create(width, height);
	GameX.AccessPixels(text_img);

	int x = 0, y = 0, i = 0;
	int val = -1;
	const char* bitsPtr = alphabetBits;
	for(;;)
	{
		char c = *bitsPtr++;
		if(!c)
			break;
		else if(c >= '0' && c <= '9')
		{
			int len = c - '0';
			while(*bitsPtr >= '0' && *bitsPtr <= '9')
			{
				len *= 10;
				len += *bitsPtr++ - '0';
			}
			while(--len)
			{
				assert(x < width && y < height);
				GameX.DrawPixel(x,y, val,val,val);
				y++;
				if(y == height)
				{
					y = 0;
					x++;
				}
			}
		}
		else
		{
			val = (c - ';') << 3;
			if(val == 248)
				val = 255;

			assert(x < width && y < height);
			GameX.DrawPixel(x,y, val,val,val);
			y++;
			if(y == height)
			{
				y = 0;
				x++;
			}
		}
	}
	GameX.EndPixelAccess();

	Initialize(text_img, 16);
}

TextPrinter::TextPrinter ()
{
	z_ownsImage = false;
	text_img = NULL;
}

TextPrinter::~TextPrinter ()
{
	if(z_ownsImage)
	{
		SAFE_DELETE(text_img);
		z_ownsImage = false;
	}
}
